#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <memory.h>
#include "phostpdk.h"
#include "explmap.h"
/* #include "pgetopt.h" */

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif
#define sgn(x) ((x) < 0 ? -1 : ((x) > 0 ? 1 : 0))

#define offsetof(struc, memb) ((size_t) ((char *) &(((struc *)0)->memb) - (char *)0))


#define  AND   &&
#define  OR    ||
#define  EQ    ==
#define  NEQ   !=
#define  GT    >
#define  GE    >=
#define  LT     <
#define  LE    =<

#ifndef Uns8
typedef unsigned char Uns8;
#endif

/* ExploreMap 2.1: extern int stricmp(const char *s1, const char *s2); */

static int
ListMatch(const char *token, const char *list)
{
    const char *tptr;
    int match;

    match=0;

    while (*list)  {
        int required = 1;
        
        for (tptr=token; *token && *list && !isspace(*list); tptr++, list++)  {
            if (required && islower(*list)) required = 0;

            if (toupper(*tptr) NEQ toupper(*list))  break;   /* Not this one */
        }
        if (*tptr == 0)  {
            if (*list && islower(*list)) required = 0;
            if (!required || isspace(*list) || *list==0) return match;
        }
        while (*list && !isspace(*list)) list++;
        if (*list) list++;
        match++;
    }
    
    return -1;
}

/*
 *                          C O N F I G U R A T I O N
 */

typedef enum {
	CFType_Uns8,
	CFType_Uns16,
	CFType_Int16,
	CFType_Uns32,
	CFType_Boolean,
    CFType_Float,
    CFType_String,
    CFType_List,
	CFType_NoType
} CFType;
typedef char NoType;

#define CFDefine(name, member, num, type, def) name,
static const char *Names[] = {
#include "explmap.hi"
0
};

#define CF_NumItems ((sizeof(Names)/sizeof(Names[0]))-1)

#define CFDefine(name, member, num, type, def) offsetof(ExplMapConfig_Struct, member),
static size_t Pos[] = {
#include "explmap.hi"
0
};

#define CFDefine(name, member, num, type, def) num,
static Uns16 Elements[] =  {
#include "explmap.hi"
0
};

#define CFDefine(name, member, num, type, def) CFType_ ## type ,
static CFType Types[] = {
#include "explmap.hi"
CFType_NoType
};

#define CFDefine(name, member, num, type, def) def,
static const char *Defaults[] = {
#include "explmap.hi"
0
};

static Boolean UserSet[CF_NumItems];

static int firstTime = 1;

static char *Data = 0;

static int readUns8(Uns16 ix, char *val);
static int readUns16(Uns16 ix, char *val);
static int readInt16(Uns16 ix, char *val);
static int readUns32(Uns16 ix, char *val);
static int readBooleanType(Uns16 ix, char *val);
static int readFloatType(Uns16 ix, char *val);
static int readListType(Uns16 ix, char *val);
static int readStringType(Uns16 ix, char *val);

/* Check each assignment for a valid parameter value. Boolean-type and List-type
   assignments need not be checked since they've already been verified by
   ListMatch */

static int
checkValue(Uns16 ix, long val)
{
	static struct {
		long minval;
		long maxval;
	} sCheck[CF_NumItems] = {
        {  1, 2000 }, /* Scan ranges */
        {  0, 2000 }, /* Detect ranges */
        {  0,  100 }, /* Defense bonus */
/* NEW: Must add the remaining configs here. They must be in the same order
        as in the .hi file */
    };
	passert(ix < CF_NumItems);
	return (val >= sCheck[ix].minval AND val <= sCheck[ix].maxval);
}

static void
doInit(void)
{
	if (!firstTime) return;
	firstTime = 0;

    Data = (char *) MemCalloc(sizeof(ExplMapConfig_Struct), 1);
}

static int
DoAssignment(const char *name, char *val)
{
	Uns16 ix;

	doInit();

	for (ix=0; ix<CF_NumItems; ix++) {
/* ExploreMap 2.1:		if (!stricmp(Names[ix], name)) break; */
	    if(!strcasecmp(Names[ix], name)) break;
	}
	if (ix >= CF_NumItems) {
		fprintf(stderr,"Parameter '%s' is not known\n",name);
		return 0;
	}

	UserSet[ix] = True;

	switch (Types[ix]) {
	 case CFType_Uns8: return readUns8(ix, val);
	 case CFType_Uns16: return readUns16(ix, val);
	 case CFType_Int16: return readInt16(ix, val);
	 case CFType_Uns32: return readUns32(ix, val);
	 case CFType_Boolean: return readBooleanType(ix, val);
     case CFType_Float: return readFloatType(ix, val);
     case CFType_String: return readStringType(ix, val);
     case CFType_List: return readListType(ix, val);
	 default:
		passert(0);
	}

	return 1;
}

static void
DoDefaultAssignments(void)
{
    int ix;
    char defstr[1024];

    doInit();

    for (ix=0; ix < CF_NumItems; ix++) {
        strcpy(defstr, Defaults[ix]);
        DoAssignment(Names[ix], defstr);
    }
    memset(UserSet, 0, sizeof(UserSet));
}

static void
PrintDefaultSets(void)
{
	int ix;

	for (ix=0; ix < CF_NumItems; ix++) {
		if (! UserSet[ix]) {
			printf("Warning: A default value will be used for parameter '%s'\n",
					Names[ix]);
		}
	}
}

static int
getLong(char *str, long *retval)
{
	char *endptr;
	char *p;

	p = strtok(str, ", \t");
	if (p==0) {
        fprintf(stderr, "Not enough elements in assignment\n");
		return 0;
	}

	*retval = strtol(p, &endptr, 0);
	if (*endptr NEQ 0) {
		fprintf(stderr,"Illegal numeric value '%s'\n", str);
		return 0;
	}
	return 1;
}

static int
getDouble(char *str, double *retval)
{
    char *endptr;
    char *p;

    p=strtok(str, ", \t");
    if (p==0) {
        fprintf(stderr,"Not enough elements in assignment");
        return 0;
    }

    *retval = strtod(p, &endptr);
    if (*endptr NEQ 0) {
        fprintf(stderr,"Illegal numeric value '%s'\n", str);
        return 0;
    }
    return 1;
}

static int
readUns8(Uns16 ix, char *val)
{
	long uval;
	Uns16 n = Elements[ix];
	Uns16 i;

	for (i=0; i<n; i++) {
		if (!getLong(val, &uval)) return 0;
		val = 0;
		if (uval < 0 OR uval > 255)  {
			fprintf(stderr, "Illegal 8-bit quantity '%ld'\n",uval);
			return 0;
		}
		if (!checkValue(ix, uval)) return 0;

		*(Uns8 *)(Data + Pos[ix] + i*sizeof(Uns8)) = (Uns8)uval;
	}
	return 1;
}

static int
readUns16(Uns16 ix, char *val)
{
	long uval;
	Uns16 n = Elements[ix];
	Uns16 i;

	for (i=0; i<n; i++) {
		if (!getLong(val, &uval)) return 0;
		val = 0;
		if (uval < 0 OR uval > 65535U)	{
			fprintf(stderr, "Illegal 16-bit unsigned quantity '%ld'\n",uval);
			return 0;
		}
		if (!checkValue(ix, uval)) return 0;

		*(Uns16 *)(Data + Pos[ix] + i*sizeof(Uns16)) = (Uns16)uval;
	}
	return 1;
}

static int
readInt16(Uns16 ix, char *val)
{
	long uval;
	Uns16 n = Elements[ix];
	Uns16 i;

	for (i=0; i<n; i++) {
		if (!getLong(val, &uval)) return 0;
		val = 0;
		if (uval < -32767 OR uval > 32767)	{
			fprintf(stderr, "Illegal 16-bit signed quantity '%ld'\n",uval);
			return 0;
		}
		if (!checkValue(ix, uval)) return 0;

		*(Int16 *)(Data + Pos[ix] + i*sizeof(Int16)) = (Int16)uval;
	}
	return 1;
}

static int
readUns32(Uns16 ix, char *val)
{
	long uval;
	Uns16 n = Elements[ix];
	Uns16 i;

	for (i=0; i<n; i++) {
		if (!getLong(val, &uval)) return 0;
		val = 0;

		if (!checkValue(ix, uval)) return 0;
		*(Uns32 *)(Data + Pos[ix] + i*sizeof(Uns32)) = (Uns32)uval;
	}
	return 1;
}

static int
readBooleanType(Uns16 ix, char *val)
{
	Uns16 n = Elements[ix];
	Uns16 i;
	char *p;

	for (i=0; i<n; i++) {
		int match;

		p = strtok(val, ", \t");
		val = 0;

		if (p == 0) {
			fprintf(stderr,"Not enough elements in assignment\n");
			return 0;
		}

		match = ListMatch(p, "False True No Yes"); /* order is important */
		if (match < 0) {
			fprintf(stderr, "Illegal boolean parameter: '%s'\n", p);
			return 0;
		}
		if (match > 1) match -= 2;	/* Convert no/yes to false/true */

		*(Boolean *)(Data + Pos[ix] + i*sizeof(Boolean)) = (Boolean)match;
	}
	return 1;
}

static int
readFloatType(Uns16 ix, char *val)
{
    double fval;
	Uns16 n = Elements[ix];
	Uns16 i;

	for (i=0; i<n; i++) {
        if (!getDouble(val, &fval)) return 0;
		val = 0;

        if (!checkValue(ix, fval)) return 0;
        *(double *)(Data + Pos[ix] + i*sizeof(double)) = fval;
	}
	return 1;
}

static int
readListType(Uns16 ix, char *val)
{
	Uns16 n = Elements[ix];
	Uns16 i;
	char *p;

	for (i=0; i<n; i++) {
		int match;

        if (! strcmp(val, Defaults[ix])) {
            match = 0;
        } else {
            p = strtok(val, ", \t");
            val = 0;

            if (p == 0) {
                fprintf(stderr,"Not enough elements in assignment\n");
                return 0;
            }

            match = ListMatch(p, Defaults[ix]);

            if (match < 0) {
                fprintf(stderr, "Illegal value: '%s'\n", p);
                return 0;
            }
        }
        *(Boolean *)(Data + Pos[ix] + i*sizeof(Boolean)) = (Boolean)match;
	}
	return 1;
}

static int
readStringType(Uns16 ix, char *val)
{
    Uns16 n = Elements[ix] - 1;

    if (strlen(val) > n) {
        fprintf(stderr,"Warning: Assignment to '%s' exceeds string length\n",
                            Names[ix]);
    }
    n = min(n, strlen(val));

    memset(Data + Pos[ix], 0, Elements[ix]);
    strncpy(Data + Pos[ix], val, n);

    return 1;
}

static char *
skipWS(char *p)
{
	while (*p && isspace(*p)) p++;
	
	if (*p == 0) return 0;
	return p;
}

static void
trimTrailingWS(char *p)
{
	char *q = p + strlen(p);
	
	while (--q >= p) {
		if (isspace(*q)) {
			*q = 0;
		} else return;
	}
}

void
ReadConfigFile(const char *pFileName)
{
    FILE *InFile = OpenInputFile(pFileName, GAME_OR_ROOT_DIR | TEXT_MODE);
    Uns16 line=0;
    char inbuf[256];

    while (!feof(InFile)) {
		char *p, *ep;
	
		line++;
        if (0 == fgets(inbuf, 255, InFile)) {
			if (!feof(InFile)) {
				fprintf(stderr,"Error reading input file at line %u\n",line);
				exit(1);
			}
			break;
		}
		
		p = skipWS(inbuf);
		if (p == 0) continue;

		if (*p == '#') continue;
		
		/* We have a name on the LHS */ 
		ep = strchr(p, '=');
		if (ep == 0) {
			fprintf(stderr,"Expected assignment at line %u\n",line);
			exit(1);
		}
		
		*ep++ = 0;
		ep = skipWS(ep);
		if (ep == 0) {
			fprintf(stderr,"Expecting value to assign at line %u\n",line);
			exit(1);
		}
		trimTrailingWS(ep);

		trimTrailingWS(p);
		if (*p == 0) {
			fprintf(stderr,"Expecting parameter name to assign to at line %u\n",line);
			exit(1);
		}

		if (!DoAssignment(p, ep)) {
            fprintf(stderr, "Parameter error at line %u\n", line);
			exit(1);
		}
	}

    PrintDefaultSets();
}



void ExplmapConfig(void)
{
    DoDefaultAssignments();
    ReadConfigFile(gConfigFileName);
    memcpy(gConfig, Data, sizeof(ExplMapConfig_Struct));
    MemFree(Data);
}
