#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <memory.h>
#include "phostpdk.h"
#include "explmap.h"

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef sqr
#define sqr(a) ((a)*(a))
#endif
#define sgn(x) ((x) < 0 ? -1 : ((x) > 0 ? 1 : 0))
#define offsetof(struc, memb) ((size_t) ((char *) &(((struc *)0)->memb) - (char *)0))

#ifndef MAXINT
#define MAXINT 0x7fff
#endif
#ifndef MAXUINT
#define MAXUINT 0xffffU
#endif


#define  AND   &&
#define  OR    ||
#define  EQ    ==
#define  NEQ   !=
#define  GT    >
#define  GE    >=
#define  LT     <
#define  LE    =<

#ifndef Uns8
typedef unsigned char Uns8;
#endif

#if 0 /* ADS: stricmp() is in the PDK for systems that don't have it.
              Besides, the implementation below has a bug in it.
      */
#ifndef __MSDOS__
int stricmp(const char* s1, const char* s2)
{
    int i=0;
    while (s1[i] AND s2[i])
    {
	if (toupper(s1[i]) NEQ toupper(s2[i])) 
	    return toupper(s1[i])-toupper(s2[i]);
	i++;
    }
    return 0;
}
#endif
#endif /* 0 */




/*****************************************/
/*       EXPLMAP variables               */
/*****************************************/


PwrapGenConfig_Struct *gPwrapConfig=NULL;
ExplMapConfig_Struct  *gConfig=NULL;

Int32          TimChecksum[2]={0xf3d19UL, 0xf4c89UL};
const Int16    MinCoordinateAllowedByPlanets=0,
               MaxCoordinateAllowedByPlanets=10000,
               DesiredDistanceFromMapEdge=0,
               gVersionMajor=2,
               gVersionMinor=5;

Uns16          gScannedPlanets[PLANET_NR+1];

const char     *gScanFileName="explmap.hst";
char           gConfigFileName[255]="explmap.cfg";

XYPlan_Structs gPlanets;





/*****************************************/
/* Miscellanous functions                */
/*****************************************/



/*
 * We generated a map and write it.
 */
void WritePlayerMap(XYPlan_Entry* pMap, Int16 pPlayer)
{
    FILE     *lMapFile;
    Uns16    i,
             bla=0;
    char     lName[16]; /* ADS: BUG: (v2.1) Changed from 11 to 16 */

    passert(pMap NEQ NULL);
    sprintf(lName,"xyplan%u.dat",pPlayer);
    lMapFile=OpenOutputFile(lName, GAME_DIR_ONLY);
    for (i=1; i<=PLANET_NR; i++)
    {
        DOSWrite16((Uns16*)pMap[i].Position, 2, lMapFile);
        DOSWrite16(&bla, 1, lMapFile);
    }
    fclose(lMapFile);
}



void CheckMap(void)
{
    Int16 i;
    Uns32 CheckX=0,
          CheckY=0;

    for (i=1; i<=PLANET_NR; i++)
    {
        if (    ((Int16)PlanetLocationX(i)<0)
             OR ((Int16)PlanetLocationY(i)<0)
           )
        {
            FreePHOSTLib();
            ErrorExit("Negative planet coordinates detected, map is corrupted.");
        }
        CheckX+=PlanetLocationX(i);
        CheckY+=PlanetLocationY(i);
    }
    if (CheckX NEQ TimChecksum[0])
    {
        FreePHOSTLib();
        Error("Map fails checksum test.");
        ErrorExit("Please correct the checksums before starting PWRAPGEN.");
    }

    if (CheckY NEQ TimChecksum[1])
    {
        Warning("Y coordinates checksum error. See EXPLMAP.DOC for more info.");
        TimChecksum[1]=CheckY;
    }
}



Boolean ReadScanFile(void)
{
    FILE *lScanFile=OpenInputFile(gScanFileName,GAME_DIR_ONLY|NO_MISSING_ERROR);
    char ReservedBytes[22];

    if (lScanFile EQ NULL)
        return False;

    if (  (!fread(ReservedBytes,22,1,lScanFile))
        OR (!DOSRead16(gScannedPlanets, PLANET_NR+1, lScanFile))
       )
    {
        fclose(lScanFile);
        return False;
    }
    fclose(lScanFile);
    return True;
}


void WriteScanFile(void)
{
    FILE *lScanFile=OpenOutputFile(gScanFileName,GAME_DIR_ONLY);
    char ReservedBytes[22];

    memset(ReservedBytes,0,22);
    ReservedBytes[0]=gVersionMajor;
    ReservedBytes[1]=gVersionMinor;
    if (    (!fwrite(ReservedBytes,22,1,lScanFile))
         OR (!DOSWrite16(gScannedPlanets, PLANET_NR+1, lScanFile))
       )
    {
        fclose(lScanFile);
        ErrorExit("Could not write explmap.hst file. Exiting.");
    }
    fclose(lScanFile);
}



/*****************************************/
/* Functions for map generating          */
/*****************************************/

Boolean PlanetIsDetected(Uns16 pId, Int16 pX, Int16 pY, RaceType_Def pScanningRace)
{
    Uns32 lScanRange,
          lDetectRange,
          lDistanceSqr,
          lDistance;

    lScanRange   =   gConfig->ScanRanges[pScanningRace]
                   * (100L-gConfig->DefenseBonus[PlanetOwner(pId)])
                   / 100;

    lDistanceSqr =  (Uns32)DistanceSquared((Int16)PlanetLocationX(pId), (Int16)PlanetLocationY(pId), pX, pY);

    if (lDistanceSqr>sqr(lScanRange))
        return False;

    if (gConfig->UseRandomScanning)
    {
        lDetectRange    =   gConfig->DetectRanges[pScanningRace]
                          * (100L-gConfig->DefenseBonus[PlanetOwner(pId)])
                          / 100;

        if (lDistanceSqr<sqr(lDetectRange))
            return True;
        /* chances : ((maxrange-distance)/(maxrange-vision))^2 */

        lDistance=sqrt(lDistanceSqr);

        if (RandomRange(1000) > (1000*sqr((float)(lScanRange-lDistance)/(float)(lScanRange-lDetectRange))))
            return False;
    }
    return True;
}



int QsortXYCompare(const void* pPtr1, const void* pPtr2)
{
    return ( ((XYPlan_Struct*)pPtr1)->X - ((XYPlan_Struct*)pPtr2)->X);
}



void MarkSensorSwept(Uns16 *pScannedPlanets)
{
    /* WARNING! Assumes no interior padding */
    struct {
        Uns16 type;
        Uns16 size;
    } header;
    Uns16 planet;
    Uns16 i;
    FILE *F;
    char utilFileName[32];

    for (i=1; i<=RACE_NR; i++) {
        if (PlayerIsActive(i)) {
            sprintf(utilFileName, "util%d.dat", i);
            F = OpenInputFile(utilFileName, GAME_DIR_ONLY | NO_MISSING_ERROR);
            if (F != NULL) {
                while (!feof(F)) {
                    DOSRead16((Uns16 *)&header, 2, F);
                    if (header.type == 6 /* Sensor Sweep */
                        || header.type == 23 /* Bioscan */
                        || (gConfig->DarkSenseLocates && header.type == 3))
                    {
                        header.size -= 2;
                        DOSRead16(&planet, 1, F);
                        pScannedPlanets[planet] |= 1 << i;
                    }
                    if (header.size)
                        fseek(F, header.size, SEEK_CUR);
                }
                fclose(F);
            }
        }
    }
}

/* The following routine checks which planets are scanned by which race. */

void DoScanning(Uns16* pScannedPlanets)
{
    Uns16 i,j,
          lScanningPlanet,
          lScanningShip,
          lOwner;

    passert(pScannedPlanets NEQ NULL);

    for (i=0; i<=PLANET_NR; i++)
    {
        if (IsPlanetExist(i)) {
            gPlanets[i].X  = PlanetLocationX(i);
            gPlanets[i].Y  = PlanetLocationY(i);
            gPlanets[i].Id = i;
        } else {
            memset(gPlanets+i, 0, sizeof(gPlanets[0]));
        }
    }

    qsort(&gPlanets[1], PLANET_NR, sizeof(XYPlan_Struct), QsortXYCompare);

    Info("\nPlanets scanning for planets ...");

    for (lScanningPlanet=1; lScanningPlanet<=PLANET_NR; lScanningPlanet++)
    {
        if (! IsPlanetExist(gPlanets[lScanningPlanet].Id)) continue;

        if ((lOwner=PlanetOwner(gPlanets[lScanningPlanet].Id)) EQ NoOwner)
            continue;

        for (j=0; j<=PLANET_NR; j++)
        {
            if (!IsPlanetExist(gPlanets[j].Id))
                continue;

            /* did we already scan the planet ? */
            if (pScannedPlanets[gPlanets[j].Id] & (1<<lOwner))
                continue;

            /* do we own the planet ? */
            if (PlanetOwner(gPlanets[j].Id) EQ lOwner)
                continue;

            if (PlanetIsDetected(gPlanets[j].Id, gPlanets[lScanningPlanet].X, gPlanets[lScanningPlanet].Y, lOwner))
                pScannedPlanets[gPlanets[j].Id] |= (1<<lOwner);
        }
    }

/* same procedure for ships */

    Info("\nShips scanning for planets ...");

    for (lScanningShip=1; lScanningShip<=SHIP_NR; lScanningShip++)
    {
        if (!IsShipExist(lScanningShip))
            continue;

        lOwner=ShipOwner(lScanningShip);

        for (j=0; j<=PLANET_NR; j++)
        {
            if (!IsPlanetExist(gPlanets[j].Id))
                continue;

            /* did we already scan the planet ? */
            if (pScannedPlanets[gPlanets[j].Id] & (1<<lOwner))
                continue;

            /* do we own the planet ? */
            if (PlanetOwner(gPlanets[j].Id) EQ lOwner)
                continue;

            if (PlanetIsDetected(gPlanets[j].Id, ShipLocationX(lScanningShip), ShipLocationY(lScanningShip), lOwner))
                pScannedPlanets[gPlanets[j].Id] |= (1<<lOwner);
        }
    }

    if (gConfig->SensorSweepsCount) {
        /* If sensor sweep finds enemy presence, we know where the planet is. */

        Info("\nMarking sensor swept planets ...");
        MarkSensorSwept(pScannedPlanets);
    }
}



void UpdateScanInfo(void)
{
    Uns16        lVisionAlliances[RACE_NR+1],
                 lPlanetAlliances[RACE_NR+1],
                 i,j,
                 lRace,
                 *lScanInfo=(Uns16*)MemAlloc(sizeof(Uns16)*(PLANET_NR+1));

    memset(lScanInfo,0,sizeof(Uns16)*(PLANET_NR+1));

    memset(lVisionAlliances,0,sizeof(lVisionAlliances));
    memset(lPlanetAlliances,0,sizeof(lPlanetAlliances));


    for (i=1; i<=11 /* NEW: RACE_NR */; i++)
    {
        lVisionAlliances[i] |= 1<<i;
        lPlanetAlliances[i] |= 1<<i;

        if (gConfig->ShareScannedPlanets)
            for (j=1; j<=11; j++)
            {
                if (    PlayerAllowsAlly(i,j,ALLY_VISION)
                    AND PlayersAreAllies(i,j)
                   )
                    lVisionAlliances[i] |= 1<<j;
            }


        if (gConfig->ShareAlliedPlanets)
            for (j=1; j<=11; j++)
            {
                if (    PlayerAllowsAlly(i,j,ALLY_PLANETS)
                    AND PlayersAreAllies(i,j)
                   )
                    lPlanetAlliances[i] |= 1<<j;
            }
    }

/* First we check for _scanned_ planets (not including own planets).
   We pass this info to allies with +v.
   Then we pass all own planets to allies with +p option.
   Finally we add all new planets to the previously scanned planets.
 */

    DoScanning(lScanInfo);

    for (i=1; i<=PLANET_NR; i++)
        if (/* NEW */ IsPlanetExist(i) AND PlanetOwner(i) NEQ NoOwner)
            gScannedPlanets[i] |= lPlanetAlliances[PlanetOwner(i)];

    for (i=1; i<=PLANET_NR; i++)
        for (lRace=Federation; lRace<=Colonies; lRace++)
            if (lScanInfo[i] & (1 << lRace))
                gScannedPlanets[i] |= lVisionAlliances[lRace];

    MemFree(lScanInfo);
}


/* A small "inline" help function : calculate checksums for planets that
 * should be on the virtual map */
void GenerateChecksumsForMapPlanets(XYPlan_Entry* pMap, Int32* pCheckX, Int32* pCheckY)
{
    Uns16 i;

    *pCheckX = *pCheckY = 0;

    for (i=1; i<=PLANET_NR; i++)
        if (!pMap[i].IsMovable)
        {
            (*pCheckX)+=pMap[i].Position[COORD_X];
            (*pCheckY)+=pMap[i].Position[COORD_Y];
        }
}

/* qsort compare */
int QsortCompare(const void* pElement1, const void* pElement2)
{
    return ((Int16*)pElement1)[0]-((Int16*)pElement2)[0];
}



void MoveSparePlanets(XYPlan_Entry* pMap, XYPlan_Element pCoord, Int32* pCheck,
                      Int32 pDesiredChecksum, Int16 pCenter)
{
    Int32  lDistanceToMove,
           lPosition,
           lCheckSum[2];
    Int16  lPlanetNr,
           lId,
           lNumber,
           lHighEdge=pCenter+gPwrapConfig->mUniverseGeometry[pCoord]/2,
           lLowEdge=pCenter-gPwrapConfig->mUniverseGeometry[pCoord]/2;

    passert((pMap NEQ NULL) AND (pCheck NEQ NULL));
    if (lHighEdge+DesiredDistanceFromMapEdge>MaxCoordinateAllowedByPlanets)
    {
        Info("Can't correct checksum. Map generation failed.");
        return;
    }

    GenerateChecksumsForMapPlanets(pMap, lCheckSum, lCheckSum+1);
    for (lId=1; lId<=PLANET_NR; lId++)
        if (pMap[lId].IsMovable)
            pMap[lId].Position[pCoord]=0;

    if (lCheckSum[pCoord]>=pDesiredChecksum)
    {
        *pCheck=lCheckSum[pCoord];
        return;
    }

    lDistanceToMove=pDesiredChecksum-lCheckSum[pCoord];
    lPlanetNr=PLANET_NR-gPwrapConfig->mNumUsablePlanets;

/* We have 2 cases.
 * (1) : We distribute the spare planets on both sides of the map.
 * (2) : All planets are moved to the high side of the map.
 *
 * We first check for case 2. If the average distance to move
 * (lDistanceToMove/lPlanetNr) is beyond the high map edge we'll simply
 * move our planets there.
 */
    lPosition=lDistanceToMove/lPlanetNr;

    if (lPosition > lHighEdge+DesiredDistanceFromMapEdge)
    {
        lPosition=min(lPosition, MaxCoordinateAllowedByPlanets);
        for (lId=1; lId<=PLANET_NR; lId++)
            if (pMap[lId].IsMovable)
            {
                pMap[lId].Position[pCoord]=lPosition;
                pMap[lId].Position[pCoord]=min(pMap[lId].Position[pCoord]+lDistanceToMove%lPlanetNr,
                                             MaxCoordinateAllowedByPlanets);
                lDistanceToMove -= pMap[lId].Position[pCoord];
                lPlanetNr--;
            }
/* Either we corrected the checksum or we moved all spare planets to the
* highest possible coordinate. */

        *pCheck=pDesiredChecksum-lDistanceToMove;
        passert(*pCheck>=0);
        return;
    }


/* Case 1 : Several subcases.
 * 1.1. We can move some planets to the far side and leave the rest at (0,0).
 * 1.2. We can move some planets to the far side and use some of the other
 *      planets to correct the checksum.
 * 1.3. We can't correct the checksum.
 */

    lNumber=lDistanceToMove/MaxCoordinateAllowedByPlanets+1;
    if (     (lNumber<=lPlanetNr)
         AND (lDistanceToMove/lNumber>lHighEdge+DesiredDistanceFromMapEdge)
       )
    {
/* case 1.1. */
        lPosition=lDistanceToMove/lNumber;
        for (lId=1; ((lId<=PLANET_NR) AND (lDistanceToMove>0)); lId++)
            if (pMap[lId].IsMovable)
            {
                pMap[lId].Position[pCoord]=lPosition;
                pMap[lId].Position[pCoord]=min(pMap[lId].Position[pCoord]+lDistanceToMove%lNumber,
                                             MaxCoordinateAllowedByPlanets);
                lDistanceToMove-=pMap[lId].Position[pCoord];
                lNumber--;
            }
        passert(lDistanceToMove EQ 0);
        *pCheck=pDesiredChecksum;
        return;
    }


/* case 1.2 and 1.3. First let's move as many planets as possible to the high
 * map regions. */
    lId=0;
    while ((lDistanceToMove>lHighEdge+DesiredDistanceFromMapEdge) AND (lId<500 /* NEW: PLANET_NR */))
    {
        lId++;
        if (pMap[lId].IsMovable)
        {
            pMap[lId].Position[pCoord]=min(lDistanceToMove, MaxCoordinateAllowedByPlanets);
            lDistanceToMove-=pMap[lId].Position[pCoord];
            lPlanetNr--;
        }
    }
    passert(lId<500);
    passert(lPlanetNr>0);

/* What can happen ? We moved as many planets as possible beyond the high edge
 * of the map. This weren't all, otherwise we would have had case 2.
 * So we have some planets left and possibly a remaining distance to move
 * in case we remaining checksum was lower than the high edge of the map.
 * Let's check it.
 */
    if (lDistanceToMove>0)
    {
        lPosition=(lDistanceToMove)/lPlanetNr;
        if (lPosition<lLowEdge-DesiredDistanceFromMapEdge)
        {
/* case 1.2 */
            for (lId++; lId<=PLANET_NR; lId++)
                if (pMap[lId].IsMovable)
                {
                    pMap[lId].Position[pCoord]=lPosition;
                    pMap[lId].Position[pCoord]=min(pMap[lId].Position[pCoord]+lDistanceToMove%lPlanetNr,
                                                   lLowEdge-DesiredDistanceFromMapEdge);
                    lDistanceToMove-=pMap[lId].Position[pCoord];
                    lPlanetNr--;
                }
            passert(lDistanceToMove EQ 0);
            *pCheck=pDesiredChecksum;
            return;
        }
        else
        {
/* case 1.3 : Checksums are still wrong. Likely cause is that the player
              scans almost all planets. In this case we give him
              the rest too. */

            Info("Map generation failed. Scanning all planets.");
            for (lId=1; lId<=PLANET_NR; lId++)
                if (pMap[lId].IsMovable)
                {
                    pMap[lId].IsMovable=0;
                    pMap[lId].Position[0]=PlanetLocationX(lId);
                    pMap[lId].Position[1]=PlanetLocationY(lId);
                }

            GenerateChecksumsForMapPlanets(pMap, lCheckSum, lCheckSum+1);
            passert(lCheckSum[pCoord] EQ pDesiredChecksum);
        }
   }
   *pCheck=pDesiredChecksum;
   return;
}



/* The following function is called if the checksums are already too big
 * after placing the regular planets. I handle only one coordinate here.
 * We have 2 possibilities : Shift the complete map from the high coordinates
 * to the low ones and place the spare planets at high values or shift just
 * some planets to get below the checksums.
 *
 * As measure of the quality of our map we take the smallest distance of the
 * desired center to the map border.
 */
void FixHighChecksums(XYPlan_Entry* pMap, XYPlan_Element pCoord, Int32* pCheck,
                      const Int32 pDesiredChecksum, Int16* pCenter)
{
    Int32         lCheck;
    Int16         lCenter1;
    XYPlan_Entry  *lTestMap=(XYPlan_Entry*)MemAlloc(sizeof(XYPlan_Entry)*(PLANET_NR+1));

    passert(pCenter NEQ NULL);
    passert(pCheck  NEQ NULL);
    passert(*pCenter>0);
    passert(*pCheck>pDesiredChecksum);

/* estimate the distance for case 1 : We switch to the high coordinates,
 * i.e. we try to center the map at the parameter given center */
    lCenter1  = *pCenter;
    memcpy(lTestMap, pMap, sizeof(XYPlan_Entry)*(PLANET_NR+1));

    MoveSparePlanets(lTestMap, pCoord, &lCheck, pDesiredChecksum, lCenter1);
    if (lCheck EQ pDesiredChecksum) /* everything perfect at the desired center */
    {
        memcpy(pMap, lTestMap, sizeof(XYPlan_Entry)*(PLANET_NR+1));
        MemFree(lTestMap);
        *pCheck=pDesiredChecksum;
        return;
    }

    ErrorExit("Map generation failure. Please zip up host data and send it to\nspock@berlin.snafu.de. Exiting.");

}



/*
 * The following procedure changes pMap so that it is the best map for
 * the given center coordinates.
 * The real center coordinates for the virtual map are returned.
 */
void GeneratePlayerMap(XYPlan_Entry* pMap, Int16 pCenter[])
{
    Int16 lCoordinate,
          lId;
    Int32 lChecksum[2]={0,0};

    for (lCoordinate=0; lCoordinate<2; lCoordinate++)
    {
/* first wrap map to the new coordinates */

        for (lId=1; lId<=PLANET_NR; lId++)
        {
            if (pMap[lId].IsMovable)
                pMap[lId].Position[lCoordinate]=0;
#if 0
            else
            {
                pMap[lId].Position[lCoordinate]+= gPwrapConfig->mUniverseGeometry[lCoordinate]*(lLowEdge/pMap[lId].Position[lCoordinate]);
                pMap[lId].Position[lCoordinate] = lLowEdge+(pMap[lId].Position[lCoordinate]-lLowEdge)%gPwrapConfig->mUniverseGeometry[lCoordinate];
            }
#endif
        }
        GenerateChecksumsForMapPlanets(pMap, lChecksum, lChecksum+1);

        if (lChecksum[lCoordinate]>TimChecksum[lCoordinate])
            FixHighChecksums(pMap, lCoordinate, &lChecksum[lCoordinate],
                            TimChecksum[lCoordinate], &pCenter[lCoordinate]);
        else
        {
             MoveSparePlanets(pMap, lCoordinate, &lChecksum[lCoordinate],
                              TimChecksum[lCoordinate], pCenter[lCoordinate]);
             if (lChecksum[lCoordinate] NEQ TimChecksum[lCoordinate])
                ErrorExit("Map generation failure. Please zip up host data and send it to\nspock@berlin.snafu.de. Exiting.");
        }
        passert(lChecksum[lCoordinate] EQ TimChecksum[lCoordinate]);
    }
}


void GenerateMaps(void)
{
    XYPlan_Entry         *lMap,
                         *lPlayerMap;
    Int16                i,
                         XHi, XLow, YHi, YLow;
    RaceType_Def         lPlayer;
    CoordinateTuple_Def  *lCenter=(CoordinateTuple_Def*)MemAlloc(sizeof(CoordinateTuple_Def)*12);

    lMap       = (XYPlan_Entry*)MemCalloc(sizeof(XYPlan_Entry),(PLANET_NR+1));
    lPlayerMap = (XYPlan_Entry*)MemAlloc(sizeof(XYPlan_Entry)*(PLANET_NR+1));

    for (i=1; i<=PLANET_NR; i++)
    {
        if (IsPlanetExist(i))
        {
            lMap[i].Position[COORD_X]=PlanetLocationX(i);
            lMap[i].Position[COORD_Y]=PlanetLocationY(i);
        }
    }


    for (lPlayer=Federation; lPlayer<=Colonies; lPlayer++)
    {

        if (PlayerIsActive(lPlayer))
        {
            Info("Generating map for player %u",lPlayer);
            XHi  = YHi  = MinCoordinateAllowedByPlanets;
            XLow = YLow = MaxCoordinateAllowedByPlanets;
            gPwrapConfig->mNumUsablePlanets=0;

            for (i=1; i<=PLANET_NR; i++)
                if (! IsPlanetExist(i))
                    lMap[i].IsMovable = 0;
                else
                {
                    if (gScannedPlanets[i] & (1<<lPlayer))
                    {
                        lMap[i].IsMovable=0;
                        XHi  = max(lMap[i].Position[0],XHi);
                        YHi  = max(lMap[i].Position[1],YHi);
                        XLow = min(lMap[i].Position[0],XLow);
                        YLow = min(lMap[i].Position[1],YLow);
                        gPwrapConfig->mNumUsablePlanets++;
                    }
                    else
                        lMap[i].IsMovable=1;
                }

#if 0
            lCenter[lPlayer][0]=XLow+(XHi-XLow)/2);
            lCenter[lPlayer][1]=YLow+(YHi-YLow)/2);
#else
            lCenter[lPlayer][0]=lCenter[lPlayer][1]=2000;
#endif
            if (gPwrapConfig->mNumUsablePlanets>0)
            {
                gPwrapConfig->mUniverseGeometry[0]=2*max(abs(2000-XHi),abs(2000-XLow));
                gPwrapConfig->mUniverseGeometry[1]=2*max(abs(2000-YHi),abs(2000-YLow));
            }
            else
                gPwrapConfig->mUniverseGeometry[0]=gPwrapConfig->mUniverseGeometry[1]=0;

            gPwrapConfig->mUniverseCenter[0]=lCenter[lPlayer][0];
            gPwrapConfig->mUniverseCenter[1]=lCenter[lPlayer][1];

            memcpy(lPlayerMap, lMap, sizeof(XYPlan_Entry)*(PLANET_NR+1));

            GeneratePlayerMap(lPlayerMap, lCenter[lPlayer]);
            WritePlayerMap(lPlayerMap, lPlayer);
        }
        else
            Info("Player %u is not active in this game.",lPlayer);
    }

    MemFree(lMap);
    MemFree(lPlayerMap);
}



void ProcessXternCommands(void)
{
    FILE    *lXternFile=OpenInputFile("xterncmd.ext",GAME_DIR_ONLY|NO_MISSING_ERROR);
    Uns16   lPlayer,
            lShare,
            lPlanetNr;
    char    *lXternCommand=(char*)MemAlloc(256),
            *lCharPtr;

    if (lXternFile EQ NULL)
        return;

    while (fgets(lXternCommand, 256, lXternFile) NEQ 0)
    {
        lCharPtr=lXternCommand;

        while (*(++lCharPtr))
            *lCharPtr=tolower(*lCharPtr);

        if ((lPlayer=atoi(lXternCommand)) EQ 0)
            continue;

        strtok(lXternCommand," \t");
        lCharPtr=strtok(NULL," \t");

        if (strncmp(lCharPtr,"shareplanets",12))
           continue;

        while ((lCharPtr=strtok(NULL," \t")) NEQ NULL)
        {
            lShare=atoi(lCharPtr);
            if ((lShare<1) OR (lShare>11))
                continue;

            for (lPlanetNr=1; lPlanetNr<=PLANET_NR; lPlanetNr++)
                if (gScannedPlanets[lPlanetNr]&(1<<lPlayer))
                    gScannedPlanets[lPlanetNr]|=(1<<lShare);
        }
    }
}


void ProcessOptions(int argc, char* argv[])
{
    static char lDirectory[2]=".";
    Uns16       lArgNr,
                lOptionNr;

    gGameDirectory = gRootDirectory = NULL;

    for (lArgNr=1; lArgNr<argc; lArgNr++)
        if (argv[lArgNr][0] NEQ '-') /* no switch */
        {
            if (gGameDirectory EQ NULL)
                gGameDirectory=argv[lArgNr];
            else
                if (gRootDirectory EQ NULL)
                    gRootDirectory=argv[lArgNr];
                else
                    ErrorExit("\nCan't process argument %s\nCall 'explmap -h' to get help\n",argv[lArgNr]);
        }
        else
            for (lOptionNr=1; lOptionNr<strlen(argv[lArgNr]); lOptionNr++)
                switch(toupper(argv[lArgNr][lOptionNr]))
                {
                    case 'H' : printf("Explore Map v%u.%u FOR PHOST >=3 ONLY!\n"
                                      "Usage : EXPLMAP [options] [Hostdir] [Rootdir]\n"
                                      "Valid options are :\n"
                                      "-h            - this help screen\n"
                                      "-v            - print version info and exit\n"
                                      "-f<filename>  - use config file <filename> instead of explmap.cfg\n",gVersionMajor, gVersionMinor);
                               exit(0);

                    case 'F' : if (strlen(&argv[lArgNr][lOptionNr+1]) EQ 0)
                                   ErrorExit("-f option : filename expected. Exiting");

                               strcpy(gConfigFileName,&argv[lArgNr][lOptionNr+1]);
                               lOptionNr=strlen(argv[lArgNr]);
                               break;

                    case 'V' : printf("This is Explore Map v%u.%u FOR PHOST >=3 ONLY!\n"
                                      "Use EXPLMAP -h to get help.\n",gVersionMajor, gVersionMinor);
                               exit(0);

                    default  : printf("Unknown option : -%c\n",argv[lArgNr][lOptionNr]);
                               break;
                }

    if (gGameDirectory EQ NULL)
        gGameDirectory = lDirectory;
    if (gRootDirectory EQ NULL)
        gRootDirectory = lDirectory;

}


int
main(int argc, char* argv[])
{
    gConfig=(ExplMapConfig_Struct*)MemAlloc(sizeof(ExplMapConfig_Struct));
    gPwrapConfig=(PwrapGenConfig_Struct*)MemAlloc(sizeof(PwrapGenConfig_Struct));
    memset(gConfig,0,sizeof(ExplMapConfig_Struct));

    ProcessOptions(argc, argv);
    SetRandomSeed((Uns16)time(0));

    Info("This is ExploreMap v%u.%u FOR PHOST >=3 ONLY!\n", gVersionMajor, gVersionMinor);
    ExplmapConfig();
    InitPHOSTLib();


    if ( (!ReadGlobalData()) OR (!ReadHostData()) )
    {
        ErrorExit("\nCan't read data files. Exiting.\n");
        FreePHOSTLib();
        exit(1);
    }

    CheckMap();
    if ( (!ReadScanFile()) OR (TurnNumber() EQ 0) )
    {
        Info("EXPLMAP.HST not found. Generating new file.");
        memset(gScannedPlanets,0,sizeof(Uns16)*(PLANET_NR+1));
    }

    ProcessXternCommands();
    UpdateScanInfo();
    GenerateMaps();

    WriteScanFile();

    FreePHOSTLib();
    MemFree(gPwrapConfig);
    MemFree(gConfig);
    return 0;
}
